#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <glib-object.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <dbus/dbus-glib.h>
#include "moblin-system-dbus.h"

enum {
	AC_ADAPTER=0,
	BATTERY,
	SUPPORTED_NAMESPACES
};

static char *namespace[SUPPORTED_NAMESPACES] = {
	"ac_adapter",
	"battery"
};

typedef struct {
	DBusGConnection *connection;
	DBusGProxy *mproxy;
	DBusGProxy *proxy[SUPPORTED_NAMESPACES];
	gchar *udi[SUPPORTED_NAMESPACES];
	GType struct_array_type;
	gboolean battery_present;
	gboolean ac_present;
	gdouble percent_charged;
	gint seconds_remaining;
	gint charge_max;
	gint charge_curr;
	gint rate;
} MoblinBattery;

gboolean         battery_dbus_connect             ();

#define HAL_DBUS_SERVICE                        "org.freedesktop.Hal"
#define HAL_DBUS_PATH_MANAGER                   "/org/freedesktop/Hal/Manager"
#define HAL_DBUS_INTERFACE_MANAGER              "org.freedesktop.Hal.Manager"
#define HAL_DBUS_INTERFACE_DEVICE               "org.freedesktop.Hal.Device"
#define g_marshal_value_peek_int(v)      g_value_get_int (v)
#define g_marshal_value_peek_string(v)   g_value_get_string (v)
#define g_marshal_value_peek_boxed(v)    g_value_get_boxed (v)

static MoblinBattery bobj;

static gboolean
GetPropertyBoolean(DBusGProxy *proxy, char *prop)
{
	GError  *error = NULL;
	gboolean value;
	gboolean ret;

     	ret = dbus_g_proxy_call (proxy, "GetPropertyBoolean", &error,
				 G_TYPE_STRING, prop,
				 G_TYPE_INVALID,
				 G_TYPE_BOOLEAN, &value,
				 G_TYPE_INVALID);

	if (error) {
		g_error_free (error);
		return FALSE;
	}

	return value;
}

static gint
GetPropertyInteger(DBusGProxy *proxy, char *prop)
{
	GError  *error = NULL;
	gint value = -1;
	gboolean ret;

     	ret = dbus_g_proxy_call (proxy, "GetPropertyInteger", &error,
				 G_TYPE_STRING, prop,
				 G_TYPE_INVALID,
				 G_TYPE_INT, &value,
				 G_TYPE_INVALID);
	if (error) {
		g_error_free (error);
		return -1;
	}

	return value;
}

#if 0
static gchar *
GetPropertyString(DBusGProxy *proxy, char *prop)
{
	GError  *error = NULL;
	gchar *value = NULL;
	gboolean ret;

     	ret = dbus_g_proxy_call (proxy, "GetPropertyString", &error,
				 G_TYPE_STRING, prop,
				 G_TYPE_INVALID,
				 G_TYPE_STRING, &value,
				 G_TYPE_INVALID);

	if (error) {
		g_error_free (error);
		return NULL;
	}

	return value;
}
#endif

void
refresh_info(MoblinSystemServer *server)
{

    if (bobj.proxy[AC_ADAPTER] != NULL) 
    {
	bobj.ac_present = 
	    GetPropertyBoolean(bobj.proxy[AC_ADAPTER], 
		"ac_adapter.present");
    }

    if (bobj.proxy[BATTERY] != NULL) 
    {
	bobj.battery_present = 
	    GetPropertyBoolean(bobj.proxy[BATTERY], 
		"battery.present");
	bobj.charge_max = 
	    GetPropertyInteger(bobj.proxy[BATTERY], 
		"battery.charge_level.last_full");
	bobj.charge_curr = 
	    GetPropertyInteger(bobj.proxy[BATTERY], 
		"battery.charge_level.current");
	bobj.rate = 
	    GetPropertyInteger(bobj.proxy[BATTERY], 
		"battery.charge_level.rate");

	if((bobj.charge_max != -1)&&(bobj.charge_curr != -1))
	{
	    bobj.percent_charged = 
		100.0*((gdouble)bobj.charge_curr/(gdouble)bobj.charge_max);
	    if(bobj.rate <= 0) {
		bobj.seconds_remaining = 0;
	    } else if(bobj.ac_present) {
		bobj.seconds_remaining = 
		    ((bobj.charge_max - bobj.charge_curr)*3600)/bobj.rate;
	    } else {
		bobj.seconds_remaining = 
		    ((bobj.charge_curr)*3600)/bobj.rate;
	    }
	}
    }
    if(server)
    {
	DBusMessage *message;
	DBusConnection *cnct;

	cnct = dbus_g_connection_get_connection(server->connection);
	message = dbus_message_new_signal(
	    "/org/moblin/SystemDaemon",
	    "org.moblin.SystemDaemon",
	    "BatteryStatusChanged");
	dbus_message_append_args(message,
	    DBUS_TYPE_BOOLEAN, &bobj.battery_present,
	    DBUS_TYPE_BOOLEAN, &bobj.ac_present,
	    DBUS_TYPE_DOUBLE, &bobj.percent_charged,
	    DBUS_TYPE_INT32, &bobj.seconds_remaining,
	    DBUS_TYPE_INVALID);

	dbus_connection_send(cnct, message, NULL);
	dbus_connection_flush(cnct);
	dbus_message_unref(message);
    }
}

gboolean
system_daemon_get_battery_status(MoblinSystemServer *server,
	gboolean *battery_attached, gboolean *ac_attached, 
	gdouble *percent_charged, gint *seconds_remaining, 
	GError **error)
{

	refresh_info(NULL);
	*battery_attached = bobj.battery_present;
	*ac_attached = bobj.ac_present;
	*percent_charged = bobj.percent_charged;
	*seconds_remaining = bobj.seconds_remaining;
	return TRUE;
}

char *
get_device_udi(DBusGProxy *proxy, char *cpb)
{
	GError *error = NULL;
	gchar **names = NULL;
	char *udi;
	int i;

	dbus_g_proxy_call (proxy, "FindDeviceByCapability", &error,
	  G_TYPE_STRING, cpb, G_TYPE_INVALID,
	  G_TYPE_STRV, &names, G_TYPE_INVALID);
	if (error) {
	    g_warning ("%s", error->message);
	    g_error_free (error);
	    return NULL;
	}
	if (names == NULL || names[0] == NULL) {
	    g_warning ("No devices of capability %s", cpb);
	    if(names) g_free(names);
	    return NULL;
	}
	udi = g_strdup (names[0]);
	for (i=0; names[i]; i++) {
	  g_free (names[i]);
	}
	g_free (names);
	return udi;
}

static void
property_modified_cb(DBusGProxy *proxy,
	gint type, GPtrArray  *properties,
	MoblinSystemServer *server)
{
	refresh_info(server);
}

static void
device_removed_cb(DBusGProxy *proxy,
	const gchar *udi,
	MoblinSystemServer *server)
{
	battery_dbus_connect();
	refresh_info(server);
}

static void
device_added_cb(DBusGProxy *proxy,
	const gchar *udi,
	MoblinSystemServer *server)
{
	battery_dbus_connect();
	refresh_info(server);
}

void
libhal_marshal_VOID__INT_BOXED (GClosure     *closure,
                                GValue       *return_value,
                                guint         n_param_values,
                                const GValue *param_values,
                                gpointer      invocation_hint,
                                gpointer      marshal_data)
{
  typedef void (*GMarshalFunc_VOID__INT_BOXED) (gpointer     data1,
                                                gint         arg_1,
                                                gpointer     arg_2,
                                                gpointer     data2);
  register GMarshalFunc_VOID__INT_BOXED callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_VOID__INT_BOXED) (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_int (param_values + 1),
            g_marshal_value_peek_boxed (param_values + 2),
            data2);
}

gboolean
battery_dbus_connect(MoblinSystemServer *server)
{
GError *error = NULL;
int i;
gboolean res=TRUE;

    if(bobj.connection == NULL)
    {
	GType struct_type;
	bobj.connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
	if(bobj.connection == NULL)
	{
	    if(error) {
		g_warning ("Could not connect to DBUS daemon: %s", 
		    error->message);
		g_error_free (error);
	    } else {
		g_warning ("Could not connect to DBUS daemon");
	    }
	    return FALSE;
	}
	bobj.mproxy = dbus_g_proxy_new_for_name (bobj.connection,
	    HAL_DBUS_SERVICE, HAL_DBUS_PATH_MANAGER, 
	    HAL_DBUS_INTERFACE_MANAGER);
	if(bobj.mproxy == NULL)
	{
	    g_warning("Failed to get proxy for %s", 
		HAL_DBUS_INTERFACE_MANAGER);
	    bobj.connection = NULL;
	    return FALSE;
	}

	dbus_g_proxy_add_signal (bobj.mproxy, "DeviceRemoved",
	    G_TYPE_STRING, G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (bobj.mproxy, "DeviceRemoved",
	    G_CALLBACK (device_removed_cb), server, NULL);
	dbus_g_proxy_add_signal (bobj.mproxy, "DeviceAdded",
	    G_TYPE_STRING, G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (bobj.mproxy, "DeviceAdded",
	    G_CALLBACK (device_added_cb), server, NULL);
	struct_type = dbus_g_type_get_struct ("GValueArray",
	    G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_INVALID);
	bobj.struct_array_type = 
	    dbus_g_type_get_collection ("GPtrArray", struct_type);
	dbus_g_object_register_marshaller (libhal_marshal_VOID__INT_BOXED,
	    G_TYPE_NONE, G_TYPE_INT, bobj.struct_array_type, G_TYPE_INVALID);
    }

    for(i = 0; i < SUPPORTED_NAMESPACES; i++)
    {
	if(bobj.udi[i] == NULL)
	{
	    bobj.udi[i] = get_device_udi(bobj.mproxy, namespace[i]);
	    if(bobj.udi[i] == NULL)
	    {
		res = FALSE;
		continue;
	    }
	}
	if (bobj.proxy[i] == NULL)
	{
	    bobj.proxy[i] = dbus_g_proxy_new_for_name(bobj.connection,
		HAL_DBUS_SERVICE, bobj.udi[i], HAL_DBUS_INTERFACE_DEVICE);
	    if (bobj.proxy[i] == NULL)
	    {
		res = FALSE;
		continue;
	    }
	    dbus_g_proxy_add_signal (bobj.proxy[i], "PropertyModified",
		G_TYPE_INT, bobj.struct_array_type, G_TYPE_INVALID);
	    dbus_g_proxy_connect_signal (bobj.proxy[i], "PropertyModified",
		G_CALLBACK (property_modified_cb), server, NULL);
	}
    }
    return res;
}

gboolean
battery_dbus_disconnect ()
{
int i;

	for(i = 0; i < SUPPORTED_NAMESPACES; i++)
	{
	    if (bobj.proxy[i] != NULL) {
		dbus_g_proxy_disconnect_signal (bobj.proxy[i], "PropertyModified",
			G_CALLBACK (property_modified_cb), &bobj);
		g_object_unref (bobj.proxy[i]);
		bobj.proxy[i] = NULL;
	    }
	}
	return TRUE;
}

gboolean
initialize_battery(MoblinSystemServer *server)
{
    int i;

    /* initialize the fields */
    bobj.connection = NULL;
    bobj.ac_present = FALSE;
    bobj.battery_present = FALSE;
    bobj.struct_array_type = G_TYPE_INVALID;
    bobj.charge_max = 0;
    bobj.charge_curr = 0;
    bobj.mproxy = NULL;
    bobj.rate = 0;
    bobj.percent_charged = 0;
    bobj.seconds_remaining = 0;
    for(i = 0; i < SUPPORTED_NAMESPACES; i++)
    {
	bobj.proxy[i] = NULL;
	bobj.udi[i] = NULL;
    }

    battery_dbus_connect(server);
    return TRUE;
}
