#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "moblin-system-dbus.h"
#include <glib/gi18n.h>

#include <sys/time.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>

// Bitmap values for the get_all_timezones interger function argument: options.
// The function can return any combination of the timezone name, latitude and
// longitude, comment, and 2-letter alphabetical code, depending on how the
// caller passes the argument as a bit-wise-or with the following bitmaps
#define TZ_ALL		16	// 0b00001111
#define TZ_CODE		8	// 0b00001000
#define TZ_LATLON	4	// 0b00000100
#define TZ_NAME		2	// 0b00000010
#define TZ_COMMENT	1	// 0b00000001

#define AVG_LINE_LEN 40	// Using a Perl script, I found the average length
						// per line in the /usr/share/zoneinfo/zone.tab file
						// was about 43 characters.  To best optimize memory
						// allocation for the get_all_timezones array of 
						// char*s, I will divide the length of the file (in 
						// characters) by the average line length (43).
						// This will return the approx of the number of lines 
						// in the zone.tab file (with 1 timezones per line)
						// so not to allocate way too many or too few char**s
						// for different (and unknown) versions of the file

// These are valid on most machines except for older
// versions of RedHat, I believe...
// TODO: add preproc directives or configure flags to check for this
#define LOCAL_INFO_DIR "/usr/share/zoneinfo/"						
#define ETC_LOCAL_DIR "/etc/localtime"
#define FILE_ZONE_TAB "/usr/share/zoneinfo/zone.tab"
			
			
extern GQuark error_quark;
enum // most error enumerations listed as <HEADERFILE>_<FUNCTION>_ERROR
{
	INVALID_INPUT,
	STDLIB_MALLOC_ERROR,
	STDLIB_REALLOC_ERROR,
	TIME_LOCALTIME_ERROR,
	TIME_MKTIME_ERROR,
	SYS_SETTIMEOFDAY_ERROR,
	SYS_GETTIMEOFDAY_ERROR,
	STDIO_FOPEN_ERROR,
	STDIO_FSEEK_FTELL_ERROR,
	STDIO_FREAD_ERROR,
	SYS_STAT_ERROR,
	UNISTD_UNLINK_ERROR,
	SYS_STAT_LSTAT_ERROR,
	UNISTD_SYMLINK_ERROR,
	UNISTD_READLINK_ERROR
};

gboolean __validate_time_input(const gint year, const gint month, const gint day, 
							 const gint hour, const gint minute, const gint second, 
							 char** errorMsg)
{
	
	// First this function checks that the caller isn't trying to set the time
	// past the year 2035 on a 32-bit machine.  Because the epoch is stored as
	// 32-bit int, the number overflows in 2038.  This may be patched on many 
	// machines but it kept failing on mine.
	if( INT_MAX <= 2147483647 ) // this is the size of an integer on a 32-bit machine
	{
		PRINTF("test\n");
		
		// If year is between 0 and 100 (YY format instead of YYYY), I add 2000
		// to it, so I have to check to make sure it isn't between 38 & 100 too.
		if( ( (year < 100) && (year >= 38) ) || (year >= 2038) )
		{
			*errorMsg = "32-bit Linux machines do not support time/dates past\n\t03:14:07 GMT Tuesday, January 19, 2038.";
			return FALSE;
		}
		
		// Even if you aren't past 2038, applications may use time values in the future
		// so for now I am not even letting you set the time within 3 years of 2038.
		if( ( (year < 100) && (year >= 35) ) || (year >= 2035) )
		{
			*errorMsg = "You are within three years of 03:14:07 GMT Tuesday, January 19, 2038.\n\tGet a 64-bit device.";
			return FALSE;
		}	
	}
	
	// If the caller messes up two inputs, I still only return the first as an
	// error.  It was easier than mallocing error and concatenating several 
	// error messages.
	if( (month > 12)  || (month < 1) )
	{
		*errorMsg = "Month must be between 1 and 12.";
		return FALSE;
	}
	
	if( (day > 31) || (day < 1) )
	{
		*errorMsg = "Day must be between 1 and 31.";
		return FALSE;
	}
	
	if( (hour > 23)  || (hour < 0) )
	{
		*errorMsg = "Hour must be between 0 and 23";
		return FALSE;
	}
	
	if( (minute > 59)  || (minute < 0) )
	{
		*errorMsg = "Minute must be between 0 and 59";
		return FALSE;
	}
	
	if( (second > 61)  || (second < 0) )
	{
		*errorMsg = "Second must be between 0 and 61";
		return FALSE;
	}
	
	return TRUE;
}

gboolean
system_daemon_set_time(MoblinSystemServer * server,
	const gint year, const gint month, const gint day,
	const gint hour, const gint minute, const gint second,
	GError **error)
{
	PRINTF("[system_daemon_set_time]:\n\tyear=%d, month=%d, day=%d\n\thour=%d, minute=%d, second=%d\n",
	year, month, day, hour, minute, second);
	
	int retVal = 0;
	
	// Validating the input
	if(!*error)
		*error = NULL;
	
	char* errorMsg = NULL;
	if( !__validate_time_input(year, month, day, hour, minute, second, &errorMsg) )
	{
		g_set_error (error, error_quark, INVALID_INPUT,
			errorMsg);
		
		PRINTF("[system_daemon_set_time] __verify_time_input failed:\n\t%s\n", errorMsg);
		return FALSE;
	}
	
	// timeval struct used to get the epoc in seconds from the time struct
	struct timeval* timev = (struct timeval*)malloc(sizeof(struct timeval));
	if( !timev )
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
			
		PRINTF("[system_daemon_set_time] malloc failed:\n\tstruct timeval* timev = NULL\n");
	
		return FALSE;	
	}
	memset(timev, 0, sizeof(struct timeval));
	
	// Timezones and Daylight Savings Time cannot be determined by using one algorithm
	// for an entire country or continent because of unpredictable political decisions in the past
	// and future which may change the zones and DST periods.  Therefore, the old method 
	// of setting timezones has been replaced by setting a symbolic link to a timezones 
	// rule file. Calls to settimeofday() should set the tz_dsttime field to NULL.
    struct timezone* timez = NULL;
    
	// time struct used to hold the time/date values passed by the caller
	struct tm* time = (struct tm*)malloc(sizeof(struct tm));
	if(!time)
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
			
		PRINTF("[system_daemon_set_time] malloc failed:\n\tstruct tm* time = NULL\n");
	
		free(timev);
		timev = NULL;
		
		return FALSE;	
	}
	memset(time, 0, sizeof(struct tm));
	
	// setting time struct members with values passed by the caller
	// with tm_isdst set to -1 (no daylight savings time set)
	// because this will be configured with the rules in the 
	// timezone file set in set_timezone
	time->tm_isdst	= -1;
	time->tm_sec	= second;
	time->tm_min 	= minute;
	time->tm_hour 	= hour;
	time->tm_mday 	= day;
	time->tm_mon	= month - 1;
		
	// if year is in format YY instead of YYYY, I am adding 
	// 2000 to make it a YYYY value
	if( (year < 100) && (year > -1) )
		time->tm_year 	= year + 2000 - 1900; // mktime expects the year - 1900
	else
		time->tm_year 	= year - 1900;
		
	// mktime returns the seconds since the Epoch given the time struct
	timev->tv_sec = mktime(time);
	if(timev->tv_sec == (time_t)(-1)) 
	{
        g_set_error (error, error_quark, TIME_MKTIME_ERROR,
			"mktime error; likely invalid input");
			
		PRINTF("[system_daemon_set_time] mktime failed:\n\tLikely that your input is not valid.\n");// errno= %s\n"), strerror(errno));
		
		free(timev);
		timev = NULL;
		
		free(time);
		time = NULL;       

		return FALSE;
	}
    PRINTF("[system_daemon_set_time] mktime successful:\n\tepoch = %d\n", ((int)timev->tv_sec));
	
	// The system_daemon_set_time function does not support milliseconds
	timev->tv_usec = 0;

	// settimeofday will set the system time to UTC given the number of 
	// seconds after the epoch
    retVal = settimeofday(timev, timez);
	if(retVal) 
	{
        g_set_error (error, error_quark, SYS_SETTIMEOFDAY_ERROR,
			strerror(errno));
		
		PRINTF("[system_daemon_set_time] settimeofday failed:\n\t%s\n", strerror(errno));
	
		free(timev);
		timev = NULL;
		
		free(time);
		time = NULL;       

		return FALSE;
	}
	PRINTF("[system_daemon_set_time] settimeofday successful:\n\ttime = %s\n", asctime (time) );
		
	free(timev);
	timev = NULL;
	
	free(time);
	time = NULL;
	
	PRINTF("[system_daemon_set_time] end\n");
	return TRUE;
}

gboolean
system_daemon_get_time(MoblinSystemServer * server, 
	gint *year, gint *month, gint *day, 
	gint *hour, gint *minute, gint *second,
	GError **error)
{
	PRINTF("[system_daemon_get_time] start\n");
	
	int retVal = 0;

    struct timeval timev;
    struct timezone timez;
	
	struct tm* getTime = NULL;

	if(!*year)
		*year = 0;
	if(!*month)
		*month = 0;
	if(!*day)
		*day = 0;
	if(!*hour)
		*hour = 0;
	if(!*minute)
		*minute = 0;
	if(!*second)
		*second = 0;
	if(!*error)
		*error = NULL;
	
	// gettimeofday will return the current system time in seconds
	// from the epoc into the timev struct
	retVal = gettimeofday(&timev, &timez);
	if(retVal) 
	{
       g_set_error (error, error_quark, SYS_GETTIMEOFDAY_ERROR,
			strerror(errno));
			
		PRINTF("[system_daemon_get_time] gettimeofday failed:\n\t%s\n", strerror(errno));
		
		return FALSE;
    }
	PRINTF("[system_daemon_get_time] gettimeofday successful:\n\tepoch = %d\n", ((int)timev.tv_sec));
	
	// localtime will convert the seconds from epoch to a time struct
	// so that we can extract the hours, minutes, secones, etc.
	getTime = localtime(&timev.tv_sec);
	if(getTime == NULL)
	{
		g_set_error (error, error_quark, TIME_LOCALTIME_ERROR,
			"localtimefailed");
			
		PRINTF("[system_daemon_get_time] localtime failed.\n");

		return FALSE;
	}
	PRINTF ( "[system_daemon_get_time] localtime successful:\n\ttime = %s\n", asctime (getTime) );
  
	// set the values to be returned to the caller
	*year	= (gint)(getTime->tm_year + 1900);
	*month 	= (gint)(getTime->tm_mon + 1);
	*day 	= (gint)(getTime->tm_mday);
	*hour 	= (gint)(getTime->tm_hour);
	*minute = (gint)(getTime->tm_min);
	*second = (gint)(getTime->tm_sec);

	PRINTF("[system_daemon_get_time] end\n");
	return TRUE;
}          

gboolean
system_daemon_get_all_timezones(MoblinSystemServer * server,
	int options, gchar ***tz_list, GError **error)
{
	PRINTF("[system_daemon_get_all_timezones] start\n\toptions=%d\n", options);
	int optionsBitMask = 8; // 0b00001000
	
	FILE* pFile = NULL;
	long fileLen;
	
	char* buffer = NULL;
		
	char* tok;
	
	char* orig = NULL;
	char* copy = NULL;
	char* origIter = NULL;
	char* copyIter = NULL;
	char* tabIter = NULL;
	
	int substrlen = 0;

	int arrLen = 0, approxLines = 0;
	int arrIter = 0;
		
	if(!options) // if options == NULL, set options to be the default which is b00000010
		options = TZ_NAME;	
	if(options > 16)
		options = TZ_ALL;
		
	// The file /usr/share/localinfo/zone.tab is a machine parsable file that
	// lists all of the available system timezones in tab deliminated columns
	pFile = fopen(FILE_ZONE_TAB, "r");
	if(pFile == NULL)
	{
		g_set_error (error, error_quark, STDIO_FOPEN_ERROR,
			strerror(errno));
			
		PRINTF("[system_daemon_get_all_timezones] fopen failed:\n\t%s\n", strerror(errno));
		
		fclose(pFile);
		pFile = NULL;		
		
		return FALSE;
	}
	PRINTF("[system_daemon_get_all_timezones] fopen successful:\n\t%s\n", strerror(errno));
	
	// getting the length of the file zone.tab
	fseek(pFile, 0, SEEK_END);
	fileLen = ftell(pFile);
	rewind(pFile);
	
	if(fileLen == 0)
	{
		g_set_error (error, error_quark, STDIO_FSEEK_FTELL_ERROR,
			"ftell/fseek error");
			
		PRINTF("[system_daemon_get_all_timezones] fseek/ftell failed:\n\t%s\n", strerror(errno));
	
		fclose(pFile);
		pFile = NULL;		
		
		return FALSE;
	}
	PRINTF("[system_daemon_get_all_timezones] fseek/ftell successful:\n\tfile length = %d\n", ((int)fileLen));
	
	// allocating a buffer to the size of the file so that we can write 
	// the file to it
	buffer = (char*)malloc(sizeof(char) * (fileLen + 1));
	if(!buffer)
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
		
		PRINTF("[system_daemon_get_all_timezones] malloc failed:\n\tchar* buffer == NULL\n");
	
		fclose(pFile);
		pFile = NULL;
		
		return FALSE;	
	}
	memset(buffer, '\0', fileLen+1);
		
	// The average length per line in the zone.tab file was about 43 characters.  
	// To best optimize memory allocation for the get_all_timezones array of 
	// char*s, I will divide the length of the file (in characters) by the average line length (43),
	// so not to allocate way too many or too few char**s for different (and unknown) versions of the file
	arrLen = approxLines = fileLen/AVG_LINE_LEN;
					
	// Allocating the number of char*s in my list of timezones
	// to be about the number of lines in the zone.tab file where
	// each line is a timezone
	(*tz_list) = (char**)malloc(sizeof(char*) * approxLines);
	if(!(*tz_list))
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
			
		PRINTF("[system_daemon_get_all_timezones] malloc failed:\n\tchar** tz_list == NULL\n");
	
		free(buffer);
		buffer = NULL;
		
		fclose(pFile);
		pFile = NULL;
		
		return FALSE;	
	}
	memset((*tz_list), 0, approxLines);

	PRINTF("[system_daemon_get_all_timezones] memory allocations successful.\n");
	
	// reading the file zone.tab into the buffer
	if (fread(buffer, sizeof(char), fileLen, pFile) != (sizeof(char) * fileLen)) 
	{
		g_set_error (error, error_quark, STDIO_FREAD_ERROR,
			strerror(errno));
			
		PRINTF("[system_daemon_get_all_timezones] fread failed:\n\t%s\n", strerror(errno)); 
		
		free(buffer);
		buffer = NULL;
		
		free(*tz_list);
		*tz_list = NULL;
		
		fclose(pFile);
		pFile = NULL;		
		
		return FALSE;
	}
	buffer[fileLen] = '\0';
	PRINTF("[system_daemon_get_all_timezones] fread successful:\n\t%ld bytes read.\n", fileLen);
	
	// tokenizing the buffer by line
	tok = strtok(buffer, "\n");
	while(tok != NULL)
	{
		optionsBitMask = 8; // 0b00001000
		
		if(tok[0] != '#') // if the line starts with a "#", it's a comment
		{
			orig = origIter = tok;  // keeping track of the mem location of beginning of the
									// char array and starting our iterator at that location
			
			// allocating a new buffer to the length of the row in the 
			// timezone table (a tab-deliminated line in zone.tab), 
			// even if we keep less than the whole line.
			// This buffer will be added to the array and passed back
			// to the caller
			copyIter = copy = (char*)malloc(sizeof(char) * (strlen(tok)+2));
			if(!copyIter)
			{
				g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
					"malloc failed");
			
				PRINTF("[system_daemon_get_all_timezones] malloc failed:\n\tchar* copyIter == NULL\n");
			
				fclose(pFile);
				pFile = NULL;
				
				return FALSE;	
			}
			memset(copyIter, '\0', strlen(tok) + 2);
						
			if(!(options & TZ_ALL)) // if we DO want the whole row in the table, skip to the
			{ 						// else and just copy it into the array
				while(optionsBitMask > 0) // The bitmask goes 8 -> 4 -> 2 -> 1 -> 0, breaking the while loop
				{
					// find the position of the next tab character in the 
					// remaining columns of the original char array
					// (walked by origIter)
					tabIter = strpbrk(origIter, "\t");
					
					// If there are no more tabs, then this row
					// may only have 3 columns
					if(!tabIter)
						tabIter = strrchr(origIter, '\0');
			
					// The length of the substring, or entry in the table, 
					// is the memory location pointed to by tabIter minus 
					// the memory location pointed to by origIter
					substrlen = tabIter - origIter + 1;
					
					if( (optionsBitMask & options) && tabIter) 	// if this is a column that the caller
					{											// wants returned, copy from origIter
						strncpy(copyIter, origIter, substrlen); // to copyIter the substring
						
						//PRINTF("origIter  = \t\"%s\"\n", origIter);
						//PRINTF("copyIter = \t\"%s\"\n", copyIter);
						
						copyIter += substrlen; // walk the copyIter char pointer forward the
						copyIter[0] = '\0';    // length of the substring
					}
					if(*tabIter == '\0') // if this row did only have three columns
						break; 			 // break out of the loop
						
					origIter += substrlen; // walk the copyIter char pointer forward the
					                       // length of the substring
										   
					optionsBitMask /= 2; // 0b00001000 -> 0b00000100 -> 0b00000010 -> 0b00000001 -> 0
				}
			} // endif(!(options & TZ_ALL))
			else // if this is a column that the caller wants returned, copy from origIter to copyIter the substring
			{    
				strncpy(copy, orig, (strlen(orig)+1));
			}
			
			// if we didn't allocate enough memory, reallocate, doubling the size
			if(arrIter > arrLen - 2)
			{
				arrLen *= 2;
				PRINTF("[system_daemon_get_all_timezones] reallocating tz_list:\n\tarrLen = %d\n", arrLen);
				
				*tz_list = (char**)realloc(*tz_list, sizeof(char*) * arrLen);
				if(!*tz_list)
				{
					g_set_error (error, error_quark, STDLIB_REALLOC_ERROR,
						"realloc failed");
				
					PRINTF("[system_daemon_get_all_timezones] realloc failed:\n\tchar** tz_list = NULL\n");
	
					free(buffer);
					buffer = NULL;
										
					fclose(pFile);
					pFile = NULL;		
					
					return FALSE;
				}
				memset((*tz_list + (sizeof(char*) * arrIter)), 0, (arrLen - arrIter));
			}
			(*tz_list)[arrIter] = copy;

			arrIter++;
		}
	
		// keep tokenizing the buffer by the newline
		tok = strtok(NULL, "\n");
	}
	
	(*tz_list)[arrIter] = NULL;
	
	free(buffer);
	buffer = NULL;
	
	fclose(pFile);
	pFile = NULL;
	
	PRINTF("[system_daemon_get_all_timezones] end\n");
	return TRUE;
}

gboolean
system_daemon_set_timezone(MoblinSystemServer * server, 
	const gchar *tz, GError **error)
{
	PRINTF("[system_daemon_set_timezone] start\n"); 
	
	//char* path = "/etc/localtime";
	char* newPath = NULL;
	int len = 0;
	 
	struct stat buffer;
	
	int retVal = 0;

    struct timeval timev;
    struct timezone timez;
	
	struct tm* gTime;
	struct tm* sTime;
	
	retVal = gettimeofday(&timev, &timez);
	if(retVal) 
	{
       g_set_error (error, error_quark, SYS_GETTIMEOFDAY_ERROR,
			strerror(errno));
			
		PRINTF("[system_daemon_set_timezone] gettimeofday failed:\n\t%s\n", strerror(errno));
		
		return FALSE;
    }
	PRINTF("[system_daemon_set_timezone] gettimeofday successful:\n\tepoch = %d\n", ((int)timev.tv_sec));
	
	gTime = localtime(&timev.tv_sec);
	if(gTime == NULL)
	{
		g_set_error (error, error_quark, TIME_LOCALTIME_ERROR,
			"localtimefailed");
			
		PRINTF("[system_daemon_set_timezone] localtime failed.\n");
		
		return FALSE;
	}
	PRINTF ( "[system_daemon_set_timezone] localtime successful:\n\ttime = %s\n", asctime (gTime) );
  
	sTime = (struct tm*)malloc(sizeof(struct tm));
	if(!sTime)
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
			
		PRINTF("[system_daemon_set_time] malloc failed:\n\tstruct tm* sTime = NULL\n");
		
		return FALSE;	
	}
	memset(sTime, 0, sizeof(struct tm));
	memcpy(sTime, gTime, sizeof(struct tm));
  
	len = strlen(tz) + strlen(LOCAL_INFO_DIR) + 1;
	newPath = (char*)malloc(sizeof(char) * len);
	if(!newPath)
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
			
		PRINTF("[system_daemon_set_timezone] malloc failed:\n\tchar* newpath = NULL\n");
	
		free(sTime);
		sTime = NULL;
			
		return FALSE;	
	}
	memset(newPath, '\0', len);
			
	strncpy(newPath, LOCAL_INFO_DIR, strlen(LOCAL_INFO_DIR));
	strncat(newPath, tz, strlen(tz));

	PRINTF("path:%s\n", newPath);
	retVal = stat(newPath, &buffer);
	if(retVal) 
	{
		if(errno != ENOENT) // lstat failed()
		{
			g_set_error (error, error_quark, SYS_STAT_ERROR,
				strerror(errno));
			
			PRINTF("[system_daemon_set_timezone] stat failed:\n\t%s\n", strerror(errno)); 
			
			free(sTime);
			sTime = NULL;
			
			free(newPath);
			newPath = NULL;
	
			return FALSE;
		}
		else // lstat() failed because the file (/etc/localtime) did not exist
		{
			g_set_error (error, error_quark, SYS_STAT_ERROR,
				"file does not exist");
				
			PRINTF("[system_daemon_set_timezone] stat failed:\n\t%s was not a valid timezone.\n", tz); 
			
			free(sTime);
			sTime = NULL;
			
			free(newPath);
			newPath = NULL;
	
			return FALSE;
		}	
	}
	
	PRINTF("[system_daemon_set_timezone] Checking for previously configured timezone.\n"); 
	
	retVal = lstat(ETC_LOCAL_DIR, &buffer);
	if(!retVal) // lstat succeeded
	{
		PRINTF("[system_daemon_set_timezone] lstat successful:\n\tPreviously configured timezone found. Removing.\n\n"); 
		
		retVal = unlink(ETC_LOCAL_DIR);
		if(retVal)
		{
			g_set_error (error, error_quark, UNISTD_UNLINK_ERROR,
				strerror(errno));
				
			PRINTF("[system_daemon_set_timezone] unlink failed:\n\t%s\n", strerror(errno)); 
			
			free(sTime);
			sTime = NULL;
			
			free(newPath);
			newPath = NULL;
	
			return FALSE;
		}
		PRINTF("[system_daemon_set_timezone] unlink successful:\n\t Previously configured timezone removed.\n"); 
	}	
	else if(errno != ENOENT) // lstat failed()
	{
		g_set_error (error, error_quark, SYS_STAT_LSTAT_ERROR,
			strerror(errno));
				
		PRINTF("[system_daemon_set_timezone] lstat failed:\n\t%s\n", strerror(errno)); 
    
		free(sTime);
		sTime = NULL;
			
		free(newPath);
		newPath = NULL;

		return FALSE;
	}
	else // lstat() failed because the file (/etc/localtime) did not exist
	{
		g_set_error (error, error_quark, SYS_STAT_LSTAT_ERROR,
			"file does not exist");
				
		PRINTF("[system_daemon_set_timezone] No previously configured timezone found.\n"); 
	}

	PRINTF("[system_daemon_set_timezone] Configuring timezone:\n\t%s\n", newPath); 
	retVal = symlink(newPath, ETC_LOCAL_DIR);
	if(retVal)
	{
		g_set_error (error, error_quark, UNISTD_SYMLINK_ERROR,
			strerror(errno));
				
		PRINTF("[system_daemon_set_timezone] symlink failed:\n\t%s\n", strerror(errno));   
		
		free(sTime);
		sTime = NULL;
			
		free(newPath);
		newPath = NULL;
			
		return FALSE;
	}
	PRINTF("[system_daemon_set_timezone] symlink successful.\n\tTimezone configured.\n"); 
	  
	// Getting the seconds since the Epoch from the time structure
	timev.tv_sec = mktime(sTime);
	if(timev.tv_sec == (time_t)(-1)) 
	{
        g_set_error (error, error_quark, TIME_MKTIME_ERROR,
			"mktime error; likely invalid input");
			
		PRINTF("[system_daemon_set_timezone] mktime failed:\n\tLikely that your input is not valid.\n");// errno= %s\n"), strerror(errno));
		
		free(sTime);
		sTime = NULL;
			
		free(newPath);
		newPath = NULL;
	
		return FALSE;
	}
    PRINTF("[system_daemon_set_timezone] mktime successful:\n\tepoch = %d\n", ((int)timev.tv_sec));

	retVal = settimeofday(&timev, &timez);
	if(retVal) 
	{
        g_set_error (error, error_quark, SYS_SETTIMEOFDAY_ERROR,
			strerror(errno));
		
		PRINTF("[system_daemon_set_timezone] settimeofday failed:\n\t%s\n", strerror(errno)); 
		
		free(sTime);
		sTime = NULL;
			
		free(newPath);
		newPath = NULL;
	
		return FALSE;
	}
	PRINTF("[system_daemon_set_timezone] settimeofday successful.\n");
		
	free(sTime);
	sTime = NULL;
		
	free(newPath);
	newPath = NULL;
			
	PRINTF("[system_daemon_set_timezone] end\n");
	return TRUE;
}

gboolean
system_daemon_get_timezone(MoblinSystemServer * server, 
	gchar **tz, GError **error)
{
	PRINTF("[system_daemon_get_timezone] start\n");
	//char* path = "/etc/localtime";
	
	struct stat buffer;
	int retVal = 0;
	int len = 0;

	char* tempTz = NULL;
	
	*tz = NULL; 
	
	PRINTF("[system_daemon_get_timezone] Checking for previously configured timezone.\n");
	retVal = lstat(ETC_LOCAL_DIR, &buffer);
	if(retVal) // lstat failed
	{
		if(errno != ENOENT) // lstat failed
		{
			g_set_error (error, error_quark, SYS_STAT_LSTAT_ERROR,
				strerror(errno));
			
			PRINTF("[system_daemon_get_timezone] lstat failed:\n\t%s\n", strerror(errno));
	        return FALSE;
		}
		else // lstat() failed because the file (/etc/localtime) did not exist
		{
			g_set_error (error, error_quark, SYS_STAT_LSTAT_ERROR,
				strerror(errno));
					
			PRINTF("[system_daemon_get_timezone] No previously configured timezone.\n");
			
			return TRUE;
		}
	} 
	
	PRINTF("[system_daemon_get_timezone] lstat successful.\n");
		
	if(!S_ISLNK(buffer.st_mode))
	{
		PRINTF("[system_daemon_get_timezone] No previously configured timezone.\n");
			
		return TRUE;
	}
		
	PRINTF("[system_daemon_get_timezone] Timezone preivoulsy configured.\n");
	
	len = (size_t)buffer.st_size + 1;
	
	tempTz = (char*)malloc(len);	
	if(!tempTz)
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
			
		PRINTF("[system_daemon_get_timezone] malloc failed:\n\tchar* tz = NULL\n");
	
		return FALSE;	
	}
	memset(tempTz, '\0', len);

	retVal = readlink(ETC_LOCAL_DIR, tempTz, len);
	if(retVal == -1)
	{
		g_set_error (error, error_quark, UNISTD_READLINK_ERROR,
			strerror(errno));
				
		PRINTF("[system_daemon_get_timezone] readlink failed:\n\t%s\n", strerror(errno));
		
		free(tempTz);
		tempTz = NULL;
		
		return FALSE;
	}
	
	if(retVal > len)
	{
		g_set_error (error, error_quark, SYS_STAT_LSTAT_ERROR,
			"readlink failed; buffer overflow");
		
		PRINTF("[system_daemon_get_timezone] readlink failed, buffer overflow:\n\t%s\n", strerror(errno));
		
		free(tempTz);
		tempTz = NULL;
		
		return FALSE;
	}
	PRINTF("[system_daemon_get_timezone] readlink successful.\n");
		
	if( tempTz != strstr(tempTz, LOCAL_INFO_DIR) )
	{
		g_set_error (error, error_quark, SYS_STAT_LSTAT_ERROR,
			"invalid timezone");
		
		PRINTF("[system_daemon_get_timezone] Timezone is invalid:\n\t%s\n", tempTz);
		
		free(tempTz);
		tempTz = NULL;
		
		return FALSE;
	}

	len = strlen(tempTz) - strlen(LOCAL_INFO_DIR) + 1;
	(*tz) = (char*)malloc(sizeof(char) * len);	
	if(!(*tz))
	{
		g_set_error (error, error_quark, STDLIB_MALLOC_ERROR,
			"malloc failed");
			
		PRINTF("[system_daemon_get_timezone] malloc failed:\n\tchar* tz = NULL\n");
		
		free(tempTz);
		tempTz = NULL;
		
		return FALSE;	
	}					
	memset((*tz), '\0', len);
	memcpy((*tz), tempTz + strlen(LOCAL_INFO_DIR), len - 1);
	(*tz)[len - 1] = '\0';
	
	PRINTF("[system_daemon_get_timezone] Get timezone successful:\n\t%s\n", *tz);
	
	free(tempTz);
	tempTz = NULL;	
		
	PRINTF("[system_daemon_get_timezone] end\n");
	return TRUE;
}

gboolean
initialize_datetime(MoblinSystemServer * server)
{
	return TRUE;
}
