/* -*- mode: c; style: linux -*- */

/* moblin-settings-brightness.c
 *
 * Written by Todd Brandt <todd.e.brandt@intel.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <glib-object.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <dbus/dbus-glib.h>
#include "moblin-system-dbus.h"
#include "lcdbacklight.h"

#define HAL_DBUS_SERVICE                 "org.freedesktop.Hal"
#define HAL_DBUS_PATH_MANAGER            "/org/freedesktop/Hal/Manager"
#define HAL_DBUS_INTERFACE_MANAGER       "org.freedesktop.Hal.Manager"
#define HAL_DBUS_INTERFACE_DEVICE        "org.freedesktop.Hal.Device"
#define HAL_DBUS_PATH_LAPTOP_PANEL       "/org/freedesktop/Hal/Device/LaptopPanel"
#define HAL_DBUS_INTERFACE_LAPTOP_PANEL  "org.freedesktop.Hal.Device.LaptopPanel"
#define HAL_DBUS_INTERFACE_KBD_BACKLIGHT "org.freedesktop.Hal.Device.KeyboardBacklight"
#define HAL_DBUS_INTERFACE_LIGHT_SENSOR  "org.freedesktop.Hal.Device.LightSensor"

typedef struct _BrightnessControl {
    DBusGProxy *proxy;
    DBusGConnection *connection;
    gchar *udi;
    guint num_levels;
    void (*brightness_changed_cb)(DBusGProxy*, guint, MoblinSystemServer*);
} BrightnessControl;

static GQuark error_quark;
enum
{
	BAD_HAL_CONNECT,
	HAL_CALL_FAILED,
	BRIGHTNESS_OUT_OF_RANGE
};

static BrightnessControl bctl;
static MoblinSystemServer *svr = NULL;

gboolean
system_daemon_get_lcdbacklight_numlevels(MoblinSystemServer *server,
	gint *num_levels, GError **error)
{
	*num_levels = bctl.num_levels;
	return TRUE;
}

gboolean
system_daemon_get_lcdbacklight_brightness(MoblinSystemServer *server,
	gint *level, GError **error)
{
	GError *derror = NULL;
	gboolean ret;
	int brightness = 0;
	BrightnessControl *bc = &bctl;

	if(bc->proxy == NULL)
	{
		g_set_error (error, error_quark, BAD_HAL_CONNECT,
			_("Bad HAL connection"));
		return FALSE;
	}

	ret = dbus_g_proxy_call (bc->proxy, "GetBrightness", &derror,
				 G_TYPE_INVALID,
				 G_TYPE_INT, &brightness,
				 G_TYPE_INVALID);
	if (derror) {
		PRINTF ("DEBUG: ERROR: %s\n", derror->message);
		g_error_free (derror);
	}
	if (ret == TRUE) {
		*level = (gint)brightness;
	} else {
		/* abort as the DBUS method failed */
		g_set_error (error, error_quark, HAL_CALL_FAILED,
			_("HAL function call failed"));
		return FALSE;
	}

	return ret;
}

gboolean
system_daemon_set_lcdbacklight_brightness(MoblinSystemServer *server,
	const gint level, GError **error)
{
	BrightnessControl *bc = &bctl;
	GError  *derror = NULL;
	gboolean ret;
	int retval;

	if((level < 0)||(level >= bc->num_levels))
	{
		g_set_error (error, error_quark, BRIGHTNESS_OUT_OF_RANGE,
			_("Brightness value out of bounds"));
		return FALSE;
	}
	if (bc->proxy == NULL)
	{
		g_set_error (error, error_quark, BAD_HAL_CONNECT,
			_("Bad HAL connection"));
		return FALSE;
	}

	ret = dbus_g_proxy_call (bc->proxy, "SetBrightness", &derror,
	           G_TYPE_INT, (int)level,
		   G_TYPE_INVALID, G_TYPE_INT, &retval, G_TYPE_INVALID);

	if (derror || ret == FALSE) {
		g_set_error (error, error_quark, HAL_CALL_FAILED,
			_("HAL function call failed"));
		g_error_free(derror);
	}

	return ret;
}

static gint
GetPropertyInteger(DBusGProxy *proxy, char *prop)
{
GError  *error = NULL;
gint value = -1;
gboolean ret;

    ret = dbus_g_proxy_call (proxy, "GetPropertyInteger", &error,
			G_TYPE_STRING, prop,
			G_TYPE_INVALID,
			G_TYPE_INT, &value,
			G_TYPE_INVALID);
    if (error) {
	PRINTF("%s\n", error->message);
	g_error_free (error);
    }
    return value;
}

static void
brightness_changed_cb (DBusGProxy *proxy,
	guint brightness, MoblinSystemServer *server)
{
DBusMessage *message;
DBusConnection *cnct;

    cnct = dbus_g_connection_get_connection(server->connection);
    message = dbus_message_new_signal(
	"/org/moblin/SystemDaemon",
	"org.moblin.SystemDaemon",
	"LCDBacklightBrightnessChanged");
    dbus_message_append_args(message,
	DBUS_TYPE_INT32, (gint *)&brightness,
	DBUS_TYPE_INVALID);
    dbus_connection_send(cnct, message, NULL);
    dbus_connection_flush(cnct);
    dbus_message_unref(message);
}

gboolean
brightness_dbus_connect (MoblinSystemServer *server, 
	BrightnessControl *bc)
{
	DBusGProxy *proxy = NULL;
	GError *error = NULL;
	gchar **names = NULL;
	int i;

	if(bc->connection == NULL)
	{
	  bc->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
	  if (error) {
	    PRINTF("WARNING: Could not connect to DBUS daemon: %s", 
	      error->message);
	    g_error_free (error);
	    bc->connection = NULL;
	    return FALSE;
	  }
	  proxy = dbus_g_proxy_new_for_name (bc->connection,
	    HAL_DBUS_SERVICE, HAL_DBUS_PATH_MANAGER, 
	    HAL_DBUS_INTERFACE_MANAGER);
	  dbus_g_proxy_call (proxy, "FindDeviceByCapability", &error,
	    G_TYPE_STRING, "laptop_panel", G_TYPE_INVALID,
	    G_TYPE_STRV, &names, G_TYPE_INVALID);
	  if (error) {
	    PRINTF("%s", error->message);
	    g_error_free (error);
	    bc->connection = NULL;
	    return FALSE;
	  }
	  if (names == NULL || names[0] == NULL) {
	    PRINTF("No devices of capability laptop_panel");
	    if(names) g_free(names);
	    bc->connection = NULL;
	    return FALSE;
	  }
	  bc->udi = g_strdup (names[0]);
	  for (i=0; names[i]; i++) {
	    g_free (names[i]);
	  }
	  g_free (names);
	}

	if (bc->proxy == NULL) {
	  proxy = dbus_g_proxy_new_for_name(bc->connection,
		HAL_DBUS_SERVICE, bc->udi,
		HAL_DBUS_INTERFACE_DEVICE);
	  if (proxy == NULL) {
	    PRINTF("Cannot connect, maybe the daemon is not running\n");
	    return FALSE;
	  }
	  bc->num_levels = 
		GetPropertyInteger(proxy, "laptop_panel.num_levels") - 1;
	  if(bc->num_levels < 0) {
	    PRINTF("no brightness levels supported\n");
	    return FALSE;
	  }

	  bc->proxy = dbus_g_proxy_new_for_name(bc->connection,
	    HAL_DBUS_SERVICE, bc->udi,
	    HAL_DBUS_INTERFACE_LAPTOP_PANEL);
	  if (bc->proxy == NULL) {
	    PRINTF("Cannot connect, maybe the daemon is not running\n");
	    return FALSE;
	  }

	  if(bc->brightness_changed_cb != NULL)
	  {
	    dbus_g_proxy_add_signal (bc->proxy, "BrightnessChanged",
	      G_TYPE_UINT, G_TYPE_INVALID);
	    dbus_g_proxy_connect_signal (bc->proxy, "BrightnessChanged",
	      G_CALLBACK (bc->brightness_changed_cb), server, NULL);
	  }
	}
	return TRUE;
}

gboolean
brightness_dbus_disconnect (BrightnessControl *bc)
{
	if (bc->proxy != NULL) {
		g_object_unref (bc->proxy);
		bc->proxy = NULL;
	}
	return TRUE;
}

gboolean
initialize_lcdbacklight(MoblinSystemServer *server)
{
    bctl.connection = NULL;
    bctl.proxy = NULL;
    bctl.udi = NULL;
    bctl.num_levels = 0;
    bctl.brightness_changed_cb = brightness_changed_cb;
    svr = server;

    error_quark = g_quark_from_string ("moblin-system-daemon-lcdbacklight");

    if(!brightness_dbus_connect(server, &bctl))
	return FALSE;

    return TRUE;
}
