/*
 * moblin-system-dbus.c
 *
 * Copyright (C) 2007 Jan Arne Petersen <jap@gnome.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 * USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include "moblin-system-dbus.h"
#include "moblin-system-marshal.h"
#ifdef HAVE_DATETIME
#include "datetime.h"
#endif
#ifdef HAVE_TOUCHSCREEN
#include "calibrator.h"
#endif
#ifdef HAVE_BATTERY
#include "battery.h"
#endif
#ifdef HAVE_LCDBACKLIGHT
#include "lcdbacklight.h"
#endif
#ifdef HAVE_SOUND
#include "sound.h"
#endif

static GObjectClass *parent_class = NULL;

G_DEFINE_TYPE (MoblinSystemServer, moblin_system_server, G_TYPE_OBJECT)
#include "moblin-system-server.h"

#ifdef HAVE_POLKIT
static DBusHandlerResult
dbus_filter_callback (DBusConnection*, DBusMessage *, void *);
#endif

GObject *
moblin_system_server_get (void)
{
	return g_object_new (MOBLIN_SYSTEM_TYPE_SERVER, NULL);
}

static GObject*
moblin_system_server_constructor (GType                  type,
				  guint                  n_construct_params,
				  GObjectConstructParam *construct_params)
{
	static GObject *singleton = NULL;
	GObject *object;

	if (!singleton) {
		singleton = G_OBJECT_CLASS (parent_class)->constructor (type, n_construct_params, construct_params);
		object = singleton;
	} else {
		object = g_object_ref (singleton);
	}

	return object;
}

static void
moblin_system_server_class_init (MoblinSystemServerClass * klass)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);

	object_class->constructor = moblin_system_server_constructor;
}

#ifdef HAVE_POLKIT
static gboolean
init_security_filter(MoblinSystemServer * server)
{
	DBusConnection *connection;

	if(server->connection == NULL) return FALSE;
	connection = dbus_g_connection_get_connection(server->connection);
        dbus_connection_setup_with_g_main(connection, NULL);
        dbus_connection_add_filter(connection,
            dbus_filter_callback, server, NULL);
	return TRUE;
}

static gboolean
init_polkit(MoblinSystemServer * server)
{
	PolKitError *p_error;
	DBusConnection *connection = dbus_g_connection_get_connection(server->connection);

	if(server->pk_context != NULL)
	    polkit_context_unref(server->pk_context);

	server->pk_context = polkit_context_new ();
	if(server->pk_context == NULL)
	{
	    PRINTF("Could not create PolicyKit context\n");
	    return FALSE;
	}

	if(server->pk_tracker != NULL)
	    polkit_tracker_unref(server->pk_tracker);

	server->pk_tracker = polkit_tracker_new ();
	if(server->pk_tracker == NULL)
	{
	    PRINTF("Could not create PolicyKit tracker\n");
	    return FALSE;
	}
	polkit_tracker_set_system_bus_connection (server->pk_tracker, connection);
	polkit_tracker_init (server->pk_tracker);
	if (!polkit_context_init (server->pk_context, &p_error))
	{
	    PRINTF("Could not init PolicyKit context: %s\n",
		polkit_error_get_error_message (p_error));
	    return FALSE;
	}
	return TRUE;
}
#endif

static gboolean
init_dbus(void *svr)
{
	MoblinSystemServer *server = (MoblinSystemServer*)svr;
	GError *error = NULL;
	DBusGProxy *driver_proxy;
	unsigned request_ret;

	if(server->connection != NULL) return TRUE;

	/* Init the DBus connection, per-server */
	server->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
	if (server->connection == NULL) {
		PRINTF("Unable to connect to dbus: %s\n",
			   error->message);
		g_error_free (error);
		return FALSE;
	}

#ifdef HAVE_POLKIT
	if(!init_polkit(server)) return FALSE;
	if(!init_security_filter(server)) return FALSE;
#endif

	/* &dbus_glib__object_info is provided in the server-bindings.h file */
	dbus_g_object_type_install_info (MOBLIN_SYSTEM_TYPE_SERVER,
					 &dbus_glib_system_daemon_object_info);

	/* Register DBUS path */
	dbus_g_connection_register_g_object (server->connection,
					     "/org/moblin/SystemDaemon",
					     G_OBJECT (server));

	/* Register the service name, the constant here are defined in dbus-glib-bindings.h */
	driver_proxy = dbus_g_proxy_new_for_name (server->connection,
						  DBUS_SERVICE_DBUS,
						  DBUS_PATH_DBUS,
						  DBUS_INTERFACE_DBUS);

	if (!org_freedesktop_DBus_request_name
	    (driver_proxy, "org.moblin.SystemDaemon", 0, &request_ret,
	     &error)) {
		PRINTF("Unable to register service: %s\n", error->message);
		g_error_free (error);
		return FALSE;
	}

	g_object_unref (driver_proxy);
	return TRUE;
}

static void
moblin_system_server_init (MoblinSystemServer * server)
{
	if(!init_dbus(server)) 
	{
		PRINTF("ERROR: Unable to establish DBUS server, the daemon will now exit, have a nice day...\n");
		exit(-1);
	}

	/* All settings libraries are initialized here */
	/* with the instance data tied to the server object */
#ifdef HAVE_DATETIME
	initialize_datetime(server);
#endif
#ifdef HAVE_TOUCHSCREEN
	initialize_touchscreen(server);
#endif
#ifdef HAVE_BATTERY
	initialize_battery(server);
#endif
#ifdef HAVE_LCDBACKLIGHT
        initialize_lcdbacklight(server);
#endif
#ifdef HAVE_SOUND
        initialize_sound(server);
#endif
}

#ifdef HAVE_POLKIT
static void
create_action(char *action, const char *iface, const char *member)
{
int i;

	sprintf(action, "%s.%s", iface, member);
	for(i = 0; i < strlen(action); i++)
	    if(isupper(action[i]))
		action[i] = tolower(action[i]);
}

static DBusHandlerResult
dbus_filter_callback (DBusConnection *connection,
        DBusMessage * message, void *svr)
{
MoblinSystemServer *server = (MoblinSystemServer *)svr;
DBusMessage *reply;
DBusError error;
PolKitError *p_error = NULL;
PolKitCaller *pk_caller = NULL;
PolKitAction *pk_action = NULL;
PolKitResult pk_result = POLKIT_RESULT_NO;
uid_t uid;
const char *sender, *member, *iface;
char action[512];

	if (dbus_message_is_signal (message, DBUS_INTERFACE_LOCAL, 
	    "Disconnected") && strcmp (dbus_message_get_path (message), 
	    DBUS_PATH_LOCAL) == 0)
	{
	    /* this is a local message; e.g. from libdbus in this process */
	    PRINTF(("Got disconnected from the system message bus; "
		"retrying to reconnect every 3000 ms"));
	    dbus_g_connection_unref(server->connection);
	    server->connection = NULL;
#ifdef HAVE_GLIB_2_14
	    g_timeout_add_seconds (3, init_dbus, server);
#else
	    g_timeout_add (3000, init_dbus, server);
#endif
	    return DBUS_HANDLER_RESULT_HANDLED;
	} else if((dbus_message_get_type(message)==DBUS_MESSAGE_TYPE_METHOD_CALL)&&
	    dbus_message_has_interface(message, "org.moblin.SystemDaemon"))
	{
	    dbus_error_init(&error);
	    sender = dbus_message_get_sender(message);
	    iface = dbus_message_get_interface(message);
	    member = dbus_message_get_member(message);
	    uid = dbus_bus_get_unix_user(connection, sender, &error);
	    if((iface != NULL)&&(member != NULL))
	    {
		create_action(action, iface, member);
	    }
	    if((sender == NULL)||(iface == NULL)||(member == NULL)) {
		PRINTF("Unidentified message received, permission denied\n");
		pk_result = POLKIT_RESULT_NO;
	    } else if((pk_action = polkit_action_new()) == NULL) {
		PRINTF("polkit_action_new failed, memory issues...\n");
		pk_result = POLKIT_RESULT_NO;
	    } else if(!polkit_action_set_action_id(pk_action, action)) {
		PRINTF("polkit_action_set_action_id failed with action %s\n", action);
		polkit_action_unref(pk_action);
		pk_result = POLKIT_RESULT_NO;
	    } else if((pk_caller = 
		polkit_tracker_get_caller_from_dbus_name(server->pk_tracker,
						sender, &error)) == NULL) {
		PRINTF("polkit_caller_new failed, memory issues...\n");
		polkit_action_unref(pk_action);
		pk_result = POLKIT_RESULT_NO;
	    } else {
		pk_result = polkit_context_is_caller_authorized(server->pk_context,
		    pk_action, pk_caller, TRUE, &p_error);
		if(polkit_error_is_set(p_error))
		{
		    PRINTF("polkit_context_is_caller_authorized error: %s - %s\n",
			polkit_error_get_error_name(p_error),
			polkit_error_get_error_message(p_error));
		    polkit_error_free(p_error);
		    pk_result = POLKIT_RESULT_NO;
		}
		polkit_action_unref(pk_action);
		polkit_caller_unref(pk_caller);
	    }

	    if(pk_result != POLKIT_RESULT_YES)
	    {
		PRINTF("CALL DENIED: iface=%s mem=%s snd=%s uid=%d\n",
		    iface, member, sender, uid);
		reply = dbus_message_new_error(message, 
		    "org.moblin.SystemDaemon.PermissionDenied",
		    "You do not have permission to call this method");
		if(reply == NULL)
		{
		    PRINTF("dbus_message_new_error failed inexplicably, probably memory issues\n");
		    exit(-1);
		}
		if(!dbus_connection_send(connection, reply, NULL))
		{
		    PRINTF("dbus_connection_send failed inexplicably, probably memory issues\n");
		    exit(-1);
		}
		dbus_message_unref(reply);
		return DBUS_HANDLER_RESULT_HANDLED;
	    }
	}

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}
#endif
