#include <config.h> 
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <termios.h>
#include <stdlib.h>
#include <sys/time.h>
#include <moblin-system-daemon/moblin-system-client.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>

#define FLAG_SWAPX              0x1
#define TST_SWAPX(x)            ((x)&FLAG_SWAPX)
#define FLAG_SWAPY              0x2
#define TST_SWAPY(x)            ((x)&FLAG_SWAPY)
#define FLAG_ROTCW              0x4
#define TST_ROTCW(x)            ((x)&FLAG_ROTCW)
#define FLAG_ROTCCW             0x8
#define TST_ROTCCW(x)           ((x)&FLAG_ROTCCW)

static DBusHandlerResult dbus_filter_callback(DBusConnection*, DBusMessage*, void*);
static DBusGConnection *connection = NULL;
static DBusGProxy *proxy = NULL;

gboolean 
activate_system_daemon (void)
{
    GError *error = NULL;

    connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
    if (connection == NULL)
    {
	g_warning("Failed to get dbus connection\n");
	g_error_free (error);
	return FALSE;
    }
    proxy = dbus_g_proxy_new_for_name (connection,
                                     "org.moblin.SystemDaemon",
                                     "/org/moblin/SystemDaemon",
                                     "org.moblin.SystemDaemon");
    if (proxy == NULL)
    {
	g_warning("Failed to establish DBUS connection\n");
	return FALSE;
    }
    return TRUE;
}

static DBusHandlerResult
dbus_filter_callback (DBusConnection * connection, DBusMessage * message, void *user_data)
{
#ifdef HAVE_TOUCHSCREEN
GError *gerror = NULL;
DBusError error;

	if (dbus_message_is_signal (message, 
		"org.moblin.SystemDaemon", "TouchScreenData"))
	{
		gint x, y;
		dbus_error_init(&error);
		if (dbus_message_get_args (message, &error,
			DBUS_TYPE_INT32, &x, DBUS_TYPE_INT32, &y,
			DBUS_TYPE_INVALID)) {
			printf("Touch Captured: x=%d, y=%d\n", x, y);
			if(!org_moblin_SystemDaemon_cancel_touch(proxy, &gerror))
			{
				g_warning ("cancel_touch error: %s", gerror->message);
				g_error_free (gerror);
			}
			exit(0);
		} else if(dbus_error_is_set(&error)) {
			dbus_error_free(&error);
		}
	}
#endif

	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

#ifdef HAVE_TOUCHSCREEN
static void
sig_handler(int num)
{
GError *gerror = NULL;

	if(num == SIGINT)
	{
		printf("Touch Captured Cancelled\n");
		if(!org_moblin_SystemDaemon_cancel_touch(proxy, &gerror))
		{
			g_warning ("cancel_touch error: %s", gerror->message);
			g_error_free (gerror);
		}
		exit(0);
	}
}
#endif

int main(int argc, char *argv[])
{
	GError *gerror = NULL;
	DBusError error;
	DBusConnection *cnct;

	if(argc < 2)
	{
		printf("Usage: %s <cmd> <args>\n", argv[0]);
		printf("Commands:\n");
#ifdef HAVE_SOUND
		printf("SOUND\n");
                printf("\tsound_num_elements\n");
                printf("\tsound_num_channels <elemidx>\n");
                printf("\tsound_element_name <elemidx>\n");
                printf("\tsound_channel_name <elemidx> <chanidx>\n");
                printf("\tsound_volume_range <elemidx>\n");
                printf("\tsound_get_volume <elemidx> <chanidx>\n");
                printf("\tsound_set_volume <elemidx> <chanidx> <vol>\n");
                printf("\tsound_set_element_volume <elemidx> <vol>\n");
                printf("\tsound_get_mute <elemidx> <chanidx>\n");
                printf("\tsound_set_mute <elemidx> <chanidx> <0or1>\n");
                printf("\tsound_set_element_mute <elemidx> <0or1>\n");
                printf("\tsound_set_all_mute <0or1>\n");
                printf("\tsound_get_playback_master\n");
                printf("\tsound_set_playback_master <elemidx>\n");
                printf("\tsound_get_record_master\n");
                printf("\tsound_set_record_master <elemidx>\n");
                printf("\tsound_capabilities <elemidx>\n");
#endif
#ifdef HAVE_DATETIME
		printf("DATE & TIME\n");
		printf("\tget_time\n");
		printf("\tset_time <MM/DD/YYYY_hh:mm:ss>\n");
		printf("\tget_timezone\n");
		printf("\tget_all_timezones\n");
		printf("\tset_timezone <region/zone>\n");
#endif
#ifdef HAVE_TOUCHSCREEN
		printf("TOUCHSCREEN\n");
		printf("\tcapture_touch\n");
		printf("\tread_calibration\n");
#endif
#ifdef HAVE_BATTERY
		printf("BATTERY\n");
		printf("\tget_battery_status\n");
#endif
#ifdef HAVE_LCDBACKLIGHT
		printf("LCDBACKLIGHT\n");
		printf("\tget_lcdbacklight_numlevels\n");
		printf("\tget_lcdbacklight_brightness\n");
		printf("\tset_lcdbacklight_brightness <value>\n");
#endif
		return 0;
	}

	g_type_init();

	if(!activate_system_daemon()) exit(-1);

	cnct = dbus_g_connection_get_connection(connection);
	if (!dbus_connection_add_filter (cnct, dbus_filter_callback, NULL, NULL)) {
		g_warning("dbus_connection_add_filter failed\n");
		return -1;
	}
	dbus_error_init(&error);
	dbus_bus_add_match (cnct,
			"type='signal',"
			"interface='org.moblin.SystemDaemon',"
			"sender='org.moblin.SystemDaemon',"
			"path='/org/moblin/SystemDaemon'", &error);
	if(dbus_error_is_set(&error)) 
	{
		g_warning("dbus_bus_add_match failed\n");
		dbus_error_free(&error);
		return -1;
	}

#ifdef HAVE_DATETIME
	if(!strcmp(argv[1], "get_time")) {
		gint year, month, day, hour, minute, second;
		if(!org_moblin_SystemDaemon_get_time(proxy,
			&year, &month, &day, &hour, &minute, &second, &gerror))
		{
		    g_warning ("get_time error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		//month++;
		printf("Date&Time = %02d/%02d/%04d_%02d:%02d:%02d\n",
			month, day, year, hour, minute, second);
	} else if(!strcmp(argv[1], "get_timezone")) {
		gchar *tz = NULL;
		if(!org_moblin_SystemDaemon_get_timezone(proxy, &tz, &gerror))
		{
		    g_warning ("get_timezone error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("Timezone = %s\n", tz);
	} else if(!strcmp(argv[1], "set_time")) {
		gint year, month, day, hour, minute, second;
		if((argc < 3)||(sscanf(argv[2], "%02d/%02d/%04d_%02d:%02d:%02d",
			&month, &day, &year, &hour, &minute, &second) != 6))
		{
			fprintf(stderr, "Usage: %s set_time MM/DD/YYYY_hh:mm:ss\n",
				argv[0]);
			return -1;
		}
		//if(month > 0) month--;
		if(!org_moblin_SystemDaemon_set_time(proxy, year, month, day, 
			hour, minute, second, &gerror))
		{
		    g_warning ("set_time error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
	} else if(!strcmp(argv[1], "set_timezone")) {
		gchar* timezone;
		if(argc < 3)
		{
			fprintf(stderr, "Usage: %s set_timezone <region/zone>\n",
				argv[0]);
			fprintf(stderr, "Use get_all_timezones to see a list of available timezones.\n");
			return -1;
		}
		timezone = g_strdup(argv[2]);
		if(!org_moblin_SystemDaemon_set_timezone(proxy, timezone, &gerror))
		{
		    g_warning ("set_timezone error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
	} else if(!strcmp(argv[1], "get_all_timezones")) {
		int optionMask = 0;
		int iter = 0;
		gchar** timezones = NULL;
		
		if(argc > 2)
		{
			if(sscanf(argv[2], "%02d", &optionMask) != 1)
			{
				fprintf(stderr, "Usage: %s get_all_timezones [0-8]\n",
					argv[0]);
				return -1;
			}
		}
		
		if(!org_moblin_SystemDaemon_get_all_timezones(proxy, optionMask, &timezones, &gerror))
		{
		    g_warning ("get_all_timezones error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		
		printf("Timezones:\n");
		while(timezones[iter])
		{
			printf("%s\n", timezones[iter]);
			
			free(timezones[iter]);
			timezones[iter] = NULL;
			
			iter++;
		}
		free(timezones);
		timezones = NULL;
	}
	
#endif
#ifdef HAVE_TOUCHSCREEN
	if(!strcmp(argv[1], "capture_touch")) {
		signal(SIGINT, sig_handler);
		if(!org_moblin_SystemDaemon_capture_touch(proxy, &gerror))
		{
		    g_warning ("capture_touch error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		while(1);
	} else if(!strcmp(argv[1], "read_calibration")) {
		gint xmin, xmax, ymin, ymax;
		guint flags;
		if(!org_moblin_SystemDaemon_read_calibration(proxy, 
			&xmin, &xmax, &ymin, &ymax, &flags, &gerror))
		{
		    g_warning ("read_calibration error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("XORG CONF: MinX=%d, MaxX=%d, MinY=%d, MaxY=%d\n",
			xmin, xmax, ymin, ymax);
		printf("           Rotate=%s, SwapX=%d, SwapY=%d\n",
			TST_ROTCW(flags)?"CW":TST_ROTCCW(flags)?"CCW":"NORMAL",
			TST_SWAPX(flags)?1:0, TST_SWAPY(flags)?1:0);
	}
#endif
#ifdef HAVE_BATTERY
	if(!strcmp(argv[1], "get_battery_status")) {
		gboolean battery_present, ac_present;
		gdouble percent_charged;
		gint seconds_remaining;
		if(!org_moblin_SystemDaemon_get_battery_status(proxy, 
			&battery_present, &ac_present, &percent_charged, 
			&seconds_remaining, &gerror))
		{
		    g_warning ("get_battery_status error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("Battery Status:\n");
		printf("\tBattery Present = %s\n", (battery_present)?"TRUE":"FALSE");
		printf("\tAC Present = %s\n", (ac_present)?"TRUE":"FALSE");
		printf("\tPercent Charged = %.2lf\n", percent_charged);
		printf("\tSeconds Remaining = %d\n", seconds_remaining);
	}
#endif
#ifdef HAVE_SOUND
	if(!strcmp(argv[1], "sound_num_elements")) {
		gint num_elements;
		if(!org_moblin_SystemDaemon_sound_num_elements(proxy, 
			&num_elements, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("%d elements found\n", num_elements);
        } else if(!strcmp(argv[1], "sound_num_channels")) {
		gint element=0, num_channels=0;
		if((argc < 3)||(sscanf(argv[2], "%d", &element) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_num_channels <elementid>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_num_channels(proxy, 
			element, &num_channels, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("%d channels found for element %d\n", 
			num_channels, element);
        } else if(!strcmp(argv[1], "sound_element_name")) {
		gint element=0;
		gchar *name;
		if((argc < 3)||(sscanf(argv[2], "%d", &element) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_element_name <elementid>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_element_name(proxy, 
			element, &name, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("element %d is named %s\n", element, name);
        } else if(!strcmp(argv[1], "sound_channel_name")) {
		gint element=0, channel=0;
		gchar *name;
		if((argc < 4)||(sscanf(argv[2], "%d", &element) != 1)||
		   (sscanf(argv[3], "%d", &channel) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_channel_name <elementid> <channelid>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_channel_name(proxy, 
			element, channel, &name, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("element %d, channel %d is named %s\n",
			element, channel, name);
        } else if(!strcmp(argv[1], "sound_volume_range")) {
		gint element=0, min, max;
		if((argc < 3)||(sscanf(argv[2], "%d", &element) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_volume_range <elementid>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_volume_range(proxy, 
			element, &min, &max, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("element %d has range [%d, %d]\n",
			element, min, max);
        } else if(!strcmp(argv[1], "sound_get_volume")) {
		gint element=0, channel=0, volume=0;
		if((argc < 4)||(sscanf(argv[2], "%d", &element) != 1)||
		   (sscanf(argv[3], "%d", &channel) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_get_volume <elementid> <channelid>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_get_volume(proxy, 
			element, channel, &volume, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("get element %d, channel %d, volume is %d\n",
			element, channel, volume);
        } else if(!strcmp(argv[1], "sound_set_volume")) {
		gint element=0, channel=0, volume=0;
		if((argc < 5)||(sscanf(argv[2], "%d", &element) != 1)||
		   (sscanf(argv[3], "%d", &channel) != 1)||
		   (sscanf(argv[4], "%d", &volume) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_set_volume <elementid> <channelid> <volume>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_set_volume(proxy, 
			element, channel, volume, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("set element %d, channel %d, to volume %d\n",
			element, channel, volume);
        } else if(!strcmp(argv[1], "sound_set_element_volume")) {
		gint element=0, volume=0;
		if((argc < 4)||(sscanf(argv[2], "%d", &element) != 1)||
		   (sscanf(argv[3], "%d", &volume) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_set_element_volume <elementid> <volume>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_set_element_volume(proxy, 
			element, volume, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("set element %d, all channels, to volume %d\n",
			element, volume);
        } else if(!strcmp(argv[1], "sound_get_mute")) {
		gint element=0, channel=0;
		gboolean mute=FALSE;
		if((argc < 4)||(sscanf(argv[2], "%d", &element) != 1)||
		   (sscanf(argv[3], "%d", &channel) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_get_mute <elementid> <channelid>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_get_mute(proxy, 
			element, channel, &mute, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("get element %d, channel %d, mute is %s\n",
			element, channel, (mute)?"ON":"OFF");
        } else if(!strcmp(argv[1], "sound_set_mute")) {
		gint element=0, channel=0, mute=0;
		if((argc < 5)||(sscanf(argv[2], "%d", &element) != 1)||
		   (sscanf(argv[3], "%d", &channel) != 1)||
		   (sscanf(argv[4], "%d", &mute) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_set_volume <elementid> <channelid> <mute 0=off/1=on>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_set_mute(proxy, 
			element, channel, (mute), &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("set element %d, channel %d, to mute %s\n",
			element, channel, (mute)?"ON":"OFF");
        } else if(!strcmp(argv[1], "sound_set_element_mute")) {
		gint element=0, mute=0;
		if((argc < 4)||(sscanf(argv[2], "%d", &element) != 1)||
		   (sscanf(argv[3], "%d", &mute) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_set_element_mute <elementid> <mute 0=off/1=on>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_set_element_mute(proxy, 
			element, (mute), &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("set element %d, all channels, to mute %s\n",
			element, (mute)?"ON":"OFF");
        } else if(!strcmp(argv[1], "sound_set_all_mute")) {
		gint mute=0;
		if((argc < 3)||(sscanf(argv[2], "%d", &mute) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_set_all_mute <mute 0=off/1=on>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_set_all_mute(proxy, 
			(mute), &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("set all elements, all channels, to mute %s\n",
			(mute)?"ON":"OFF");
        } else if(!strcmp(argv[1], "sound_get_playback_master")) {
		gint element;
		if(!org_moblin_SystemDaemon_sound_get_playback_master(proxy, 
			&element, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("playback master is %d\n", element);
        } else if(!strcmp(argv[1], "sound_set_playback_master")) {
		gint element;
		if((argc < 3)||(sscanf(argv[2], "%d", &element) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_set_playback_master <element>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_set_playback_master(proxy, 
			element, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("playback master is %d\n", element);
        } else if(!strcmp(argv[1], "sound_get_record_master")) {
		gint element;
		if(!org_moblin_SystemDaemon_sound_get_record_master(proxy, 
			&element, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("record master is %d\n", element);
        } else if(!strcmp(argv[1], "sound_set_record_master")) {
		gint element;
		if((argc < 3)||(sscanf(argv[2], "%d", &element) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_set_record_master <element>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_set_record_master(proxy, 
			element, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("record master is %d\n", element);
        } else if(!strcmp(argv[1], "sound_capabilities")) {
		gint element=0;
		gboolean volume, mute, playback;
		if((argc < 3)||(sscanf(argv[2], "%d", &element) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s sound_capabilities <elementid>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_sound_capabilities(proxy, 
			element, &volume, &mute, &playback, &gerror))
		{
		    g_warning ("error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("element %d: %s volume=%s, mute=%s\n",
			element, (playback)?"Playback":"Record",
			(volume)?"YES":"NO", (mute)?"YES":"NO");
	}
#endif
#ifdef HAVE_LCDBACKLIGHT
	if(!strcmp(argv[1], "get_lcdbacklight_numlevels")) {
		gint numlevels;
		if(!org_moblin_SystemDaemon_get_lcdbacklight_numlevels(proxy, 
			&numlevels, &gerror))
		{
		    g_warning ("get_lcdbacklight_numlevels error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("Number of LCD Backlight levels: %d\n", numlevels);
	} else if(!strcmp(argv[1], "get_lcdbacklight_brightness")) {
		gint level;
		if(!org_moblin_SystemDaemon_get_lcdbacklight_brightness(proxy, 
			&level, &gerror))
		{
		    g_warning ("get_lcdbacklight_brightness error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
		printf("Current LCD Backlight level: %d\n", level);
	} else if(!strcmp(argv[1], "set_lcdbacklight_brightness")) {
		gint level;
		if((argc < 3)||(sscanf(argv[2], "%d", &level) != 1))
		{
		    fprintf(stderr, 
			"Usage: %s set_lcdbacklight_brightness <level>\n",
			argv[0]);
		    return -1;
		}
		if(!org_moblin_SystemDaemon_set_lcdbacklight_brightness(proxy, 
			level, &gerror))
		{
		    g_warning ("set_lcdbacklight_brightness error: %s", gerror->message);
		    g_error_free (gerror);
		    return -1;
		}
	}
#endif
	return 0;
}
