/* mixer.c
 *
 * Written by Todd Brandt <todd.e.brandt@intel.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <glib-object.h>
#include <math.h>
#include <fcntl.h>
#include <sys/types.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <alsa/asoundlib.h>
#include <sys/poll.h>
#include "sound.h"

#define ALSADEVFILE "/dev/snd/seq"

enum {
	ELEM_NONE=0,
	ELEM_COMMON,
	ELEM_PLAYBACK,
	ELEM_CAPTURE
};

enum {
	DUP_NAMES_IGNORED,
	DUP_NAMES_ALLOWED
};

struct ptrack {
        /* static info */
        gchar *name;
        snd_mixer_selem_id_t *sid;
        snd_mixer_elem_t *elem;
        char has_vol_ctl;
        char has_switch_ctl;
        long vmin;
        long vmax;
	int num_channels;
	int *channels;
};

static GQuark error_quark;
enum
{
	INVALID_ELEMENT_IDX,
	INVALID_CHANNEL_IDX,
	ELEMENT_HAS_NO_VOLUME,
	ELEMENT_HAS_NO_MUTE
};

static struct snd_mixer_selem_regopt smixer_options;
static char card[64] = "default";

/* Master list of all available playback tracks from the sound card */
static struct ptrack *ptlist = NULL;
static int ptnum = 0;
static int playback_master = -1;
static int record_master = -1;
static int init = 0;

gboolean
system_daemon_sound_num_elements(
        MoblinSystemServer *server,
        gint *num_elements,
        GError **error)
{
	if(num_elements) *num_elements = ptnum;
	return TRUE;
}

gboolean
system_daemon_sound_num_channels(
        MoblinSystemServer *server,
        const gint e,
        gint *num_channels,
        GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}

	if(num_channels) *num_channels = t->num_channels;
	return TRUE;
}

gboolean
system_daemon_sound_get_playback_master(
        MoblinSystemServer *server,
        gint *element,
        GError **error)
{
	if(element) *element = playback_master;
	return TRUE;
}

gboolean
system_daemon_sound_set_playback_master(
        MoblinSystemServer *server,
        const gint e,
        GError **error)
{
	if((e < 0)||(e >= ptnum))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	playback_master = e;
	return TRUE;
}

gboolean
system_daemon_sound_get_record_master(
        MoblinSystemServer *server,
        gint *element,
        GError **error)
{
	if(element) *element = record_master;
	return TRUE;
}

gboolean
system_daemon_sound_set_record_master(
        MoblinSystemServer *server,
        const gint e,
        GError **error)
{
	if((e < 0)||(e >= ptnum))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	record_master = e;
	return TRUE;
}

gboolean
system_daemon_sound_element_name(
        MoblinSystemServer *server,
        const gint e,
        gchar **name,
        GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if(name) *name = g_strdup(t->name);
	return TRUE;
}

gboolean
system_daemon_sound_channel_name(
        MoblinSystemServer *server,
        const gint e,
        const gint c,
        gchar **name,
        GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	    ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if((c < 0)||(c >= t->num_channels))
	{
		g_set_error (error, error_quark, INVALID_CHANNEL_IDX,
			_("Channel index is out of range"));
		return FALSE;
	}
	if(name) *name = g_strdup(snd_mixer_selem_channel_name(t->channels[c]));
	return TRUE;
}

gboolean
system_daemon_sound_volume_range(
        MoblinSystemServer *server,
        const gint e,
        gint *min,
        gint *max,
        GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if(!t->has_vol_ctl)
	{
		g_set_error (error, error_quark, ELEMENT_HAS_NO_VOLUME,
			_("Element has no volume control"));
		return FALSE;
	}
	if(min) *min = t->vmin;
	if(max) *max = t->vmax;

	return TRUE;
}

gboolean
system_daemon_sound_capabilities(
        MoblinSystemServer *server,
        const gint e,
        gboolean *volume,
        gboolean *mute,
        gboolean *playback,
	GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}

	if(volume) *volume = (t->has_vol_ctl)?TRUE:FALSE;
	if(mute) *mute = (t->has_switch_ctl)?TRUE:FALSE;
	if((t->has_vol_ctl==ELEM_CAPTURE)||
		(t->has_switch_ctl==ELEM_CAPTURE))
	{
		*playback = FALSE;
	} else {
		*playback = TRUE;
	}

	return TRUE;
}

gboolean
system_daemon_sound_get_volume(
        MoblinSystemServer *server,
        const gint e,
        const gint c,
        gint *volume,
        GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if((c < 0)||(c >= t->num_channels))
	{
		g_set_error (error, error_quark, INVALID_CHANNEL_IDX,
			_("Channel index is out of range"));
		return FALSE;
	}
	if(!t->has_vol_ctl)
	{
		g_set_error (error, error_quark, ELEMENT_HAS_NO_VOLUME,
			_("Element has no volume control"));
		return FALSE;
	}

	if(t->has_vol_ctl == ELEM_CAPTURE) {
	    snd_mixer_selem_get_capture_volume(t->elem,
		t->channels[c], (long *)volume);
	} else {
	    snd_mixer_selem_get_playback_volume(t->elem,
		t->channels[c], (long *)volume);
	}

	return TRUE;
}

gboolean
system_daemon_sound_set_volume(
        MoblinSystemServer *server,
        const gint e,
        const gint c,
        const gint volume,
        GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if((c < 0)||(c >= t->num_channels))
	{
		g_set_error (error, error_quark, INVALID_CHANNEL_IDX,
			_("Channel index is out of range"));
		return FALSE;
	}
	if(!t->has_vol_ctl)
	{
		g_set_error (error, error_quark, ELEMENT_HAS_NO_VOLUME,
			_("Element has no volume control"));
		return FALSE;
	}

	if(t->has_vol_ctl == ELEM_CAPTURE) {
	    snd_mixer_selem_set_capture_volume(t->elem,
		t->channels[c], volume);
	} else {
	    snd_mixer_selem_set_playback_volume(t->elem,
		t->channels[c], volume);
	}

	return TRUE;
}

gboolean
system_daemon_sound_set_element_volume(
        MoblinSystemServer *server,
        const gint e,
        const gint volume,
        GError **error)
{
	struct ptrack *t;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if(!t->has_vol_ctl)
	{
		g_set_error (error, error_quark, ELEMENT_HAS_NO_VOLUME,
			_("Element has no volume control"));
		return FALSE;
	}

	if(t->has_vol_ctl == ELEM_CAPTURE) {
	    snd_mixer_selem_set_capture_volume_all(t->elem, volume);
	} else {
	    snd_mixer_selem_set_playback_volume_all(t->elem, volume);
	}

	return TRUE;
}

gboolean
system_daemon_sound_get_mute(
        MoblinSystemServer *server,
        const gint e,
        const gint c,
        gboolean *mute,
        GError **error)
{
	struct ptrack *t;
	int val = 0;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if((c < 0)||(c >= t->num_channels))
	{
		g_set_error (error, error_quark, INVALID_CHANNEL_IDX,
			_("Channel index is out of range"));
		return FALSE;
	}
	if(!t->has_switch_ctl)
	{
		g_set_error (error, error_quark, ELEMENT_HAS_NO_MUTE,
			_("Element has no mute control"));
		return FALSE;
	}

	if(t->has_switch_ctl == ELEM_CAPTURE) {
	    snd_mixer_selem_get_capture_switch(t->elem,
		t->channels[c], &val);
	} else {
	    snd_mixer_selem_get_playback_switch(t->elem,
		t->channels[c], &val);
	}
	*mute = (val)?TRUE:FALSE;

	return TRUE;
}

gboolean
system_daemon_sound_set_mute(
        MoblinSystemServer *server,
        const gint e,
        const gint c,
        const gboolean mute,
        GError **error)
{
	struct ptrack *t;
	int val = 0;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if((c < 0)||(c >= t->num_channels))
	{
		g_set_error (error, error_quark, INVALID_CHANNEL_IDX,
			_("Channel index is out of range"));
		return FALSE;
	}
	if(!t->has_switch_ctl)
	{
		g_set_error (error, error_quark, ELEMENT_HAS_NO_MUTE,
			_("Element has no mute control"));
		return FALSE;
	}

	val = (mute)?1:0;
	if(t->has_switch_ctl == ELEM_CAPTURE) {
	    snd_mixer_selem_set_capture_switch(t->elem,
		t->channels[c], val);
	} else {
	    snd_mixer_selem_set_playback_switch(t->elem,
		t->channels[c], val);
	}

	return TRUE;
}

gboolean
system_daemon_sound_set_element_mute(
        MoblinSystemServer *server,
        const gint e,
        const gboolean mute,
        GError **error)
{
	struct ptrack *t;
	int val = 0;

	if((e < 0)||(e >= ptnum)||!ptlist||!init||
	  ((t=&ptlist[e]) == NULL))
	{
		g_set_error (error, error_quark, INVALID_ELEMENT_IDX,
			_("Element index is out of range"));
		return FALSE;
	}
	if(!t->has_switch_ctl)
	{
		g_set_error (error, error_quark, ELEMENT_HAS_NO_MUTE,
			_("Element has no mute control"));
		return FALSE;
	}

	val = (mute)?1:0;
	if(t->has_switch_ctl == ELEM_CAPTURE) {
	    snd_mixer_selem_set_capture_switch_all(t->elem, val);
	} else {
	    snd_mixer_selem_set_playback_switch_all(t->elem, val);
	}

	return TRUE;
}

gboolean
system_daemon_sound_set_all_mute(
        MoblinSystemServer *server,
        const gboolean mute,
        GError **error)
{
	struct ptrack *t;
	int e, val = 0;

	if(!ptlist||!init) return TRUE;

	val = (mute)?1:0;
	for(e = 0; e < ptnum; e++)
	{
	    if(((t = &ptlist[e]) == NULL)||!t->has_switch_ctl)
		continue;
	    if((t->has_switch_ctl == ELEM_PLAYBACK)||
		(t->has_switch_ctl == ELEM_COMMON))
	    {
		snd_mixer_selem_set_playback_switch_all(t->elem, val);
	    }
	}
	return TRUE;
}

void process_elements(snd_mixer_t *handle, int etype, int allow_dups)
{
	int sw, i, playback=(etype==ELEM_PLAYBACK);
	long vol;
	snd_mixer_elem_t *elem;
	snd_mixer_selem_id_t *sid;
	char *t, message[1000], name[100];
	const char *hwname;
	int has_vol_ctl, has_switch_ctl, index;

	snd_mixer_selem_id_alloca(&sid);

	/* we test every element on the sound card */
	for (elem = snd_mixer_first_elem(handle); elem; 
		elem = snd_mixer_elem_next(elem))
	{
		snd_mixer_selem_get_id(elem, sid);
		hwname = snd_mixer_selem_id_get_name(sid);
		PRINTF("Found hw element %s\n", hwname);
		if (!snd_mixer_selem_is_active(elem)) {
			PRINTF("Disregarding %s, inactive\n", hwname);
			continue;
		}
		/* we need to see what volume and swtich controls */
		/* each element has, because we need at least one */
		has_vol_ctl = ELEM_NONE;
		has_switch_ctl = ELEM_NONE;

		if(playback) {
		    if (snd_mixer_selem_has_common_volume(elem)) {
			has_vol_ctl = ELEM_COMMON;
		    } else if (snd_mixer_selem_has_playback_volume(elem)) {
			has_vol_ctl = ELEM_PLAYBACK;
		    }
		    if (snd_mixer_selem_has_common_switch(elem)) {
			has_switch_ctl = ELEM_COMMON;
		    } else if (snd_mixer_selem_has_playback_switch(elem)) {
			has_switch_ctl = ELEM_PLAYBACK;
		    }
		} else {
		    if (!snd_mixer_selem_has_common_volume(elem)&&
			snd_mixer_selem_has_capture_volume(elem))
		    {
			has_vol_ctl = ELEM_CAPTURE;
		    }
		    if (!snd_mixer_selem_has_common_switch(elem)&&
			snd_mixer_selem_has_capture_switch(elem))
		    {
			has_switch_ctl = ELEM_CAPTURE;
                    }
		}

		/* if the elem has no volume or switch capability, ignore it */
		if((has_vol_ctl == ELEM_NONE)&&(has_switch_ctl == ELEM_NONE)) {
			PRINTF("Disregarding %s, has no controls\n", hwname);
			continue;
		}
		/* element name from hardware */
		hwname = snd_mixer_selem_id_get_name(sid);
		/* element index */
		index = snd_mixer_selem_id_get_index(sid);

		/* if this is not the first index, it's a dup */
		if(index&&!allow_dups) {
			PRINTF("Disregarding %s, duplicate\n", hwname);
			continue;
		}
		if(index) {
			sprintf(name, "%s%i", hwname, index);
		} else {
			strcpy(name, hwname);
		} 

		/* at this point we want the element, so add a space for */
		/* it in the list */
		if(ptlist == NULL) {
			ptlist = (struct ptrack*)malloc(
				(++ptnum)*sizeof(struct ptrack));
		} else {
			ptlist = (struct ptrack*)realloc(ptlist, 
				(++ptnum)*sizeof(struct ptrack));
		}
		memset(&ptlist[ptnum-1], 0, sizeof(struct ptrack));

		/* initialize what we already know about this element */
		ptlist[ptnum-1].name = g_strdup(name);
		ptlist[ptnum-1].sid = sid;
		ptlist[ptnum-1].elem = elem;
		ptlist[ptnum-1].has_vol_ctl = has_vol_ctl;
		ptlist[ptnum-1].has_switch_ctl = has_switch_ctl;

		/* determine the element's volume range */
		if(has_vol_ctl)
		{
		    if(playback) {
			snd_mixer_selem_get_playback_volume_range(elem, 
				&ptlist[ptnum-1].vmin, &ptlist[ptnum-1].vmax);
		    } else {
			snd_mixer_selem_get_capture_volume_range(elem, 
				&ptlist[ptnum-1].vmin, &ptlist[ptnum-1].vmax);
		    }
		}

		ptlist[ptnum-1].num_channels = 0;
		ptlist[ptnum-1].channels = NULL;
		for(i = 0; i < SND_MIXER_SCHN_LAST; i++)
		{
		    if((playback&&snd_mixer_selem_has_playback_channel(elem, i))||
		       (!playback&&snd_mixer_selem_has_capture_channel(elem, i)))
		    {
			if(ptlist[ptnum-1].channels == NULL) {
			    ptlist[ptnum-1].channels = 
				(int*)malloc(
				(++(ptlist[ptnum-1].num_channels))*sizeof(int));
			} else {
			    ptlist[ptnum-1].channels = 
				(int*)realloc(ptlist[ptnum-1].channels, 
				(++(ptlist[ptnum-1].num_channels))*sizeof(int));
			}
			ptlist[ptnum-1].channels[ptlist[ptnum-1].num_channels-1] = i;
		    }
		}

		sprintf(message, "Found Element %s:", name);
		t = message+strlen(message);

		for(i = 0; i < ptlist[ptnum-1].num_channels; i++)
		{
		    sprintf(t, " %s",
			snd_mixer_selem_channel_name(ptlist[ptnum-1].channels[i]));
		    t += strlen(t);
		    if(has_vol_ctl)
		    {
			if(playback) {
			    snd_mixer_selem_get_playback_volume(elem, 
				ptlist[ptnum-1].channels[i], &vol);
			} else {
			    snd_mixer_selem_get_capture_volume(elem, 
				ptlist[ptnum-1].channels[i], &vol);
			}
			sprintf(t, "-V%d", (int)vol);
			t += strlen(t);
		    }
		    if(has_switch_ctl)
		    {
			if(playback) {
			    snd_mixer_selem_get_playback_switch(elem, 
				ptlist[ptnum-1].channels[i], &sw);
			} else {
			    snd_mixer_selem_get_capture_switch(elem, 
				ptlist[ptnum-1].channels[i], &sw);
			}
			sprintf(t, "-S%d", sw);
			t += strlen(t);
		    }
		}

		PRINTF("%s\n", message);

		if(playback&&has_vol_ctl&&has_switch_ctl)
			playback_master = ptnum-1;

		if(!playback&&has_vol_ctl&&has_switch_ctl)
			record_master = ptnum-1;
	}
	if(playback&&(playback_master < 0)) playback_master = ptnum-1;
	if(!playback&&(record_master < 0)) record_master = ptnum-1;
}

gboolean
initialize_sound(MoblinSystemServer *server)
{
	int err;
	snd_mixer_t *handle = NULL;

	if(init) return 0;

	error_quark = g_quark_from_string ("moblin-system-daemon-sound");

	if(ptlist != NULL) {
		free(ptlist);
		ptlist = NULL;
		ptnum = 0;
	}

	if(access(ALSADEVFILE, F_OK)) {
		PRINTF("Can not access the Alsa device node!");
		return -1;
	}

	smixer_options.device = card;

 	if ((err = snd_mixer_open(&handle, 0)) < 0) {
		PRINTF("Mixer %s open error: %s", card, snd_strerror(err));
		handle = NULL;
		return -1;
	}

	if ((err = snd_mixer_attach(handle, card)) < 0) {
		PRINTF("Mixer attach %s error: %s", card, snd_strerror(err));
		snd_mixer_close(handle);
		handle = NULL;
		return -1;
	}
	if ((err = snd_mixer_selem_register(handle, NULL, NULL)) < 0)
	{
		PRINTF("Mixer register error: %s", snd_strerror(err));
		snd_mixer_close(handle);
		handle = NULL;
		return -1;
	}
	err = snd_mixer_load(handle);
	if (err < 0) {
		PRINTF("Mixer %s load error: %s", card, snd_strerror(err));
		snd_mixer_close(handle);
		handle = NULL;
		return -1;
	}

	/* playback & common elements */
	process_elements(handle, ELEM_PLAYBACK, DUP_NAMES_ALLOWED);
	/* capture elements */
	process_elements(handle, ELEM_CAPTURE, DUP_NAMES_IGNORED);

	if(ptnum < 1) {
		PRINTF("No usable sound elements found\n");
		return -1;
	}
	init = 1;
	return 0;
}
